# -*- coding: utf-8 -*-
import re
import sys
import textwrap

import semver

from suite_py.lib import logger
from suite_py.lib.handler import prompt_utils
from suite_py.lib.handler.drone_handler import DroneHandler
from suite_py.lib.handler.git_handler import GitHandler
from suite_py.lib.handler.github_handler import GithubHandler


class BatchJob:
    def __init__(
        self, project, config, tokens, environment, cpu_request, memory_request
    ):
        self._project = project
        self._config = config
        self._tokens = tokens
        self._environment = environment
        self._cpu_request = cpu_request
        self._memory_request = memory_request
        self._git = GitHandler(project, config)
        self._github = GithubHandler(tokens)
        self._drone = DroneHandler(config, tokens, repo=project)
        self._countries = _parse_available_countries(self._drone)
        self._repo = self._github.get_repo(project)

    def run(self):
        country = prompt_utils.ask_choices(
            "Which country do you want to run the job on?",
            self._countries,
        )

        if not _has_job_pipeline(self._drone, country, self._environment):
            logger.error(
                "There is no job pipeline for the specified country/environment"
            )
            sys.exit(1)

        command = prompt_utils.ask_questions_input(
            "What command do you want to execute?", f"/app/bin/{self._project} "
        )

        self._start_job(country, command)

    def _start_job(self, country, command):
        if self._environment == "staging":
            build = self._get_latest_master_build()

            self._ask_confirm(
                country,
                command,
                f'Current master: {build["message"]} (#{build["number"]})',
            )

            promotion = self._drone.promote_staging(
                build["number"],
                f"job-{country}-{self._environment}",
                f"BATCH_COMMAND={command}&JOB_CPU={self._cpu_request}&JOB_MEMORY={self._memory_request}",
            )
        elif self._environment == "production":
            version = self._get_latest_tag(country)

            self._ask_confirm(country, command, f"Current tag: {version}")

            promotion = self._drone.promote_production(
                version,
                f"job-{country}-{self._environment}",
                f"DRONE_TAG={version}&BATCH_COMMAND={command}&JOB_CPU={self._cpu_request}&JOB_MEMORY={self._memory_request}",
            )

        if "number" not in promotion:
            logger.warning(f"Unable to promote drone build. Response: {promotion}")
            return

        logger.info("Drone build started successfully!")
        logger.info(
            f"You can follow the build status here: {self._drone.get_build_url(promotion['number'])}"
        )

    def _get_latest_master_build(self):
        try:
            builds = self._drone.get_builds_from_branch("master")

            latest_build = None
            non_green_builds = []
            for b in builds:
                if b["status"] == "success":
                    if latest_build is None:
                        latest_build = b
                    break

                non_green_builds.append(b)

            if len(non_green_builds) > 0:
                logger.warning(
                    "There are recent builds on master still running or failed"
                )
                for b in non_green_builds:
                    logger.warning(
                        f'* {b["message"]} (Status: {b["status"]}, Number: #{b["number"]})'
                    )

            if not latest_build:
                logger.error("Unable to find latest build on master")
                sys.exit(255)

            return latest_build

        except Exception as e:
            print(e)
            logger.error(
                "An error has occured retrieving current master version during batch job.\nPlease ask #team-devops for help."
            )
            sys.exit(255)

    def _get_latest_tag(self, country):
        try:
            logger.info("Retrieving latest version, this may take some time...")
            # get latest 10 tags
            tags = self._github.get_tags(self._project)
            #  exclude tags that don't match semver notation
            semver_tags = [t for t in tags if semver.VersionInfo.isvalid(t.name)][0:9]

            builds = []
            for tag in semver_tags:
                for b in self._drone.get_builds_from_tag(tag.name):
                    if (
                        b["event"] == "promote"
                        and b["status"] == "success"
                        and b["deploy_to"] == f"deploy-{country}-production"
                        and "params" in b
                        and "DRONE_TAG" in b["params"]
                    ):
                        builds.append(b)

            if not builds:
                # may end up here if there isn't any successfull build for latest 10 tags
                logger.error(
                    "An error has occured retrieving current version during rollback.\nPlease ask #team-devops for help."
                )
                sys.exit(255)

            # get latest build using build number as key
            current_build = max(builds, key=lambda x: x["number"])

            if not current_build:
                logger.error(
                    f"Unable to determine current version for country {country}"
                )
                sys.exit(255)

            v = semver.VersionInfo.parse(current_build["params"]["DRONE_TAG"])
            logger.info(f"Current version for country {country}: {v}")

            return v
        except Exception as e:
            print(e)
            logger.error(
                "An error has occured retrieving current version during batch job.\nPlease ask #team-devops for help."
            )
            sys.exit(255)

    def _ask_confirm(self, country, command, message):
        logger.info(
            textwrap.dedent(
                f"""
                You're about to run a batch job on project {self._project}, for env {self._environment}, in country {country}.
                {message}
                Command: {command}
                """
            )
        )

        if not prompt_utils.ask_confirm("Do you confirm?", default=False):
            sys.exit(0)


def _parse_available_countries(drone):
    pipelines = drone.parse_yaml()

    if pipelines is None:
        logger.error("The file .drone.yml was not found. Unable to continue.")
        sys.exit(1)

    countries = []
    REGEX = re.compile(r"deploy-([a-z]+)-.*")
    for pipeline in pipelines:
        if "name" in pipeline:
            c = REGEX.findall(pipeline["name"])
            if len(c) > 0 and c[0] is not None and c[0] not in countries:
                countries.append(c[0])

    return countries


def _has_job_pipeline(drone, country, environment):
    pipelines = drone.parse_yaml()

    if pipelines is None:
        logger.error("The file .drone.yml was not found. Unable to continue.")
        sys.exit(1)

    job_pipeline = f"job-{country}-{environment}"
    for pipeline in pipelines:
        if "name" in pipeline and pipeline["name"] == job_pipeline:
            return True

    return False



from requests.auth import AuthBase


class BearerAuth(AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["authorization"] = "Bearer " + self.token
        return r



# -*- coding: utf-8 -*-
import re
from datetime import datetime
from pathlib import Path
from typing import Iterator, Match, Optional, Tuple

DEFAULT_CHANGELOG_PATH = "CHANGELOG.md"


class ChangelogHandler:
    """
    Manages changelog-related stuff across different commands.
    """

    TAG_REGEX = r"^## \[([0-9]+\.[0-9]+\.[0-9]+)\]( - \d+-\d+-\d+)?"
    UNRELEASED_HEADER_REGEX = r"^## \[Unreleased\]\s*"
    UNRELEASED_VERSION_LINK_REGEX = r"^\[Unreleased\]:\s*.+"

    def __init__(self, changelog_path=DEFAULT_CHANGELOG_PATH) -> None:
        self.changelog_path = Path(changelog_path)

    def bump_changelog(self, repo_html_url, current_version, next_version):
        """
        Updates the changelog by adding the new version header at the top and related link at the bottom.

        Args:
            `repo_html_url`: Url of the repo "for humans" i.e. no the API one
            `current_version`: Version before bump
            `next_version`: Version after bump
        """
        changelog_text = self.changelog_path.read_text("utf-8")
        changelog_text = self._update_version_header(changelog_text, next_version)
        changelog_text = self._update_version_links(
            changelog_text, repo_html_url, current_version, next_version
        )
        self.changelog_path.write_text(changelog_text, "utf-8")

    def get_latest_entry_with_tag(self):
        """
        Gets the latest changelog entry with the respective tag.

        Returns:
            A tuple with `(latest_tag, changelog)`.
        """
        changelog_text = self.changelog_path.read_text("utf-8")
        try:
            tag, content = next(self.get_sections(changelog_text))
            return tag, self._replace_h3_with_h1(content)
        except ValueError:
            return None, None

    def get_sections(self, changelog_text: str) -> Iterator[Tuple[str, str]]:
        """
        Returns an iterator over sections i.e. (tag, content) pairs.
        """
        if not re.search(self.TAG_REGEX, changelog_text, flags=re.MULTILINE):
            raise ValueError("No entries available in the CHANGELOG")

        tags = list(re.finditer(self.TAG_REGEX, changelog_text, flags=re.MULTILINE))
        tags = tags if len(tags) > 1 else [tags[0], None]

        yield from (
            self._extract_tags_and_content(prev_tag, next_tag, changelog_text)
            for prev_tag, next_tag in self._pairwise(tags)
        )

    def changelog_exists(self) -> bool:
        """
        Checks whether the CHANGELOG exists or not.
        """
        return self.changelog_path.exists()

    def _extract_tags_and_content(
        self, prev_tag: Match[str], next_tag: Optional[Match[str]], changelog_text: str
    ):
        tag = prev_tag.group(1)
        content_end = -1

        if next_tag is None:
            start_of_version_links = re.search(
                self.UNRELEASED_VERSION_LINK_REGEX, changelog_text, flags=re.MULTILINE
            )
            content_end = start_of_version_links.start()
        else:
            content_end = next_tag.start()

        content = changelog_text[prev_tag.end() : content_end].strip()
        return tag, content

    def _pairwise(self, l: list):
        yield from zip(l, l[1:])

    def _replace_h3_with_h1(self, line: str) -> str:
        return re.sub("^###", "#", line, flags=re.MULTILINE)

    def _update_version_header(self, changelog_text: str, next_version: str) -> str:
        today = datetime.today().strftime("%Y-%m-%d")
        new_version_header = (
            "## [Unreleased]\n\n" "---\n\n" f"## [{next_version}] - {today}\n\n"
        )
        return re.sub(
            self.UNRELEASED_HEADER_REGEX,
            new_version_header,
            changelog_text,
            flags=re.MULTILINE,
        )

    def _update_version_links(
        self,
        changelog_text: str,
        repo_html_url: str,
        current_version: str,
        next_version: str,
    ) -> str:
        if (
            re.search(
                self.UNRELEASED_VERSION_LINK_REGEX, changelog_text, flags=re.MULTILINE
            )
            is not None
        ):
            new_version_links = (
                f"\n[Unreleased]: {repo_html_url}/compare/{next_version}...HEAD"
                f"\n[{next_version}]: {repo_html_url}/compare/{current_version}...{next_version}"
            )

            return re.sub(
                self.UNRELEASED_VERSION_LINK_REGEX,
                new_version_links,
                changelog_text,
                flags=re.MULTILINE,
            )

        starting_version_links = (
            f"\n[Unreleased]: {repo_html_url}/compare/{next_version}...HEAD"
            f"\n[{next_version}]: {repo_html_url}/releases/tag/{next_version}\n"
        )
        return changelog_text + starting_version_links



# -*- encoding: utf-8 -*-
import inquirer
from InquirerPy import prompt


def ask_confirm(text, default=True):
    questions = [
        {"type": "confirm", "message": text, "name": "continue", "default": default}
    ]
    answer = prompt(questions)
    return bool(answer["continue"])


def ask_questions_input(question_text, default_text=""):
    questions = [
        {
            "type": "input",
            "name": "value",
            "message": question_text,
            "default": default_text,
        }
    ]

    answer = prompt(questions)
    return answer["value"]


def ask_questions_editor(question_text, default_text=""):
    questions = [
        inquirer.Editor(
            "value",
            message=question_text,
            default=default_text,
        )
    ]

    answer = inquirer.prompt(questions)
    return answer["value"]


def ask_choices(question_text, choices, default_text=""):
    questions = [
        {
            "type": "list",
            "name": "value",
            "message": question_text,
            "choices": choices,
            "default": default_text,
        }
    ]

    answer = prompt(questions)
    return answer["value"]


def ask_multiple_choices(question_text, choices):
    questions = [
        {
            "type": "checkbox",
            "name": "values",
            "message": question_text,
            "choices": [
                {"name": choice, "value": choice, "enabled": False}
                for choice in choices
            ],
        }
    ]

    answer = prompt(questions)
    return answer["values"]


def checkbox_with_manual_fallback(
    checkbox_msg, fallback_msg, choices, autocomplete=None
):

    question = inquirer.Checkbox(
        "value",
        checkbox_msg,
        choices,
        other=True,
        autocomplete=autocomplete,
    )
    answer = inquirer.prompt([question])["value"]
    if not answer:
        question = inquirer.Text(
            "value",
            fallback_msg,
            autocomplete=autocomplete,
        )
        answer = [inquirer.prompt([question])["value"]]
    return answer


def make_completer(comprehension):
    """
    Builds am inquirer completer for the given list generator. The generator
    should consume a string and return a list of strings for auto-complete suggestion
    """

    def _completer(text, state):
        options = comprehension(text)
        try:
            return options[state]
        except IndexError:
            return None

    return _completer
